# -*- coding: utf-8; -*-
"""
Copyright (C) 2018 Mario Aichinger <aichingm@gmail.com>
Copyright (C) 2007-2012 Lincoln de Sousa <lincoln@minaslivre.org>
Copyright (C) 2007 Gabriel Falcão <gabrielteratos@gmail.com>

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License as
published by the Free Software Foundation; either version 2 of the
License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
General Public License for more details.

You should have received a copy of the GNU General Public
License along with this program; if not, write to the
Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
Boston, MA 02110-1301 USA
"""
import enum
import logging
import os
import re
import subprocess
import time
import yaml

import cairo

import gi

gi.require_version("Gtk", "3.0")
gi.require_version("Gdk", "3.0")

from gi.repository import Gdk
from gi.repository import Gtk
from guake.globals import ALIGN_BOTTOM
from guake.globals import ALIGN_CENTER
from guake.globals import ALIGN_LEFT
from guake.globals import ALIGN_RIGHT
from guake.globals import ALIGN_TOP

try:
    from gi.repository import GdkX11
except ImportError:
    GdkX11 = False

log = logging.getLogger(__name__)


def gdk_is_x11_display(instance):
    if GdkX11:
        return isinstance(instance, GdkX11.X11Display)
    return False


def get_server_time(widget):
    try:
        return GdkX11.x11_get_server_time(widget.get_window())
    except (TypeError, AttributeError):
        # Issue: https://github.com/Guake/guake/issues/1071
        # Wayland does not seem to like `x11_get_server_time`.
        # Use local timestamp instead
        ts = time.time()
        return ts


# Decorator for save-tabs-when-changed
def save_tabs_when_changed(func):
    """Decorator for save-tabs-when-changed"""

    def wrapper(*args, **kwargs):
        # Find me the Guake!
        clsname = args[0].__class__.__name__
        g = None
        if clsname == "Guake":
            g = args[0]
        elif getattr(args[0], "get_guake", None):
            g = args[0].get_guake()
        elif getattr(args[0], "get_notebook", None):
            g = args[0].get_notebook().guake
        elif getattr(args[0], "guake", None):
            g = args[0].guake
        elif getattr(args[0], "notebook", None):
            g = args[0].notebook.guake

        func(*args, **kwargs)
        log.debug("mom, I've been called: %s %s", func.__name__, func)

        # Tada!
        if g and g.settings.general.get_boolean("save-tabs-when-changed"):
            g.save_tabs()

    return wrapper


def save_preferences(filename):
    # XXX: Hardcode?
    prefs = subprocess.check_output(["dconf", "dump", "/org/guake/"])
    with open(filename, "wb") as f:
        f.write(prefs)


def restore_preferences(filename):
    # XXX: Hardcode?
    with open(filename, "rb") as f:
        prefs = f.read()
    with subprocess.Popen(["dconf", "load", "/org/guake/"], stdin=subprocess.PIPE) as p:
        p.communicate(input=prefs)


class FileManager:
    def __init__(self, delta=1.0):
        self._cache = {}
        self._delta = max(0.0, delta)

    def clear(self):
        self._cache.clear()

    def read_yaml(self, filename: str):

        content = None

        try:
            content = self.read(filename)
        except PermissionError:
            log.debug("PermissionError while reading %s.", filename)
        except FileNotFoundError:
            log.debug("File %s does not exists.", filename)
        except UnicodeDecodeError:
            log.debug("Encoding error %s (we assume is utf-8).", filename)

        if content is not None:
            try:
                content = yaml.safe_load(content)
            except yaml.YAMLError:
                log.debug("YAMLError reading %s.", filename)
                content = None
        return content

    def read(self, filename: str) -> str:
        # Return the content of a file from the fs or from cache.
        if (
            filename not in self._cache
            or self._cache[filename]["time"] + self._delta < time.monotonic()
        ):
            with open(filename, mode="r", encoding="utf-8") as fd:
                self._cache[filename] = {"time": time.monotonic(), "content": fd.read()}
        return self._cache[filename]["content"]


class TabNameUtils:
    @classmethod
    def shorten(cls, text, settings):
        use_vte_titles = settings.general.get_boolean("use-vte-titles")
        if not use_vte_titles:
            return text
        max_name_length = settings.general.get_int("max-tab-name-length")
        if max_name_length != 0 and len(text) > max_name_length:
            text = "..." + text[-max_name_length:]
        return text


class HidePrevention:
    def __init__(self, window):
        """Create a new HidePrevention object like `HidePrevention(window)`"""
        if not isinstance(window, Gtk.Window):
            raise ValueError(f"window must be of type Gtk.Window, not of type {type(window)}")
        self.window = window

    def may_hide(self):
        """returns True if the window is allowed to hide and
        False if `prevent()` is called from some where
        """
        return getattr(self.window, "can_hide", True)

    def prevent(self):
        """sets a flag on the window object which indicates to
        may_hide that the window is NOT allowed to be hidden.
        """
        setattr(self.window, "can_hide", False)

    def allow(self):
        """sets the flag so that it indicates to may_hide that the window is allowed to be hidden"""
        setattr(self.window, "can_hide", True)


class FullscreenManager:
    FULLSCREEN_ATTR = "is_fullscreen"

    def __init__(self, settings, window, guake=None):
        self.settings = settings
        self.window = window
        self.guake = guake
        self.window_state = None

    def is_fullscreen(self):
        return getattr(self.window, self.FULLSCREEN_ATTR, False)

    def set_window_state(self, window_state):
        self.window_state = window_state
        setattr(
            self.window,
            self.FULLSCREEN_ATTR,
            bool(window_state & Gdk.WindowState.FULLSCREEN),
        )

        if not window_state & Gdk.WindowState.WITHDRAWN:
            if self.is_fullscreen():
                self.fullscreen()
            elif window_state & Gdk.WindowState.FOCUSED and self.guake.hidden:
                self.unfullscreen()

    def fullscreen(self):
        self.window.fullscreen()
        setattr(self.window, self.FULLSCREEN_ATTR, True)
        self.toggle_fullscreen_hide_tabbar()

    def unfullscreen(self):
        self.window.unfullscreen()
        setattr(self.window, self.FULLSCREEN_ATTR, False)
        self.toggle_fullscreen_hide_tabbar()

        # FIX to unfullscreen after show, fullscreen, hide, unfullscreen
        # (unfullscreen breaks/does not shrink window size)
        RectCalculator.set_final_window_rect(self.settings, self.window)

    def toggle(self):
        if self.is_fullscreen():
            self.unfullscreen()
        else:
            self.fullscreen()

    def toggle_fullscreen_hide_tabbar(self):
        if self.is_fullscreen():
            if (
                self.settings.general.get_boolean("fullscreen-hide-tabbar")
                and self.guake
                and self.guake.notebook_manager
            ):
                self.guake.notebook_manager.set_notebooks_tabbar_visible(False)
        else:
            if self.guake and self.guake.notebook_manager:
                v = self.settings.general.get_boolean("window-tabbar")
                self.guake.notebook_manager.set_notebooks_tabbar_visible(v)


class RectCalculator:
    @classmethod
    def set_final_window_rect(cls, settings, window):
        """Sets the final size and location of the main window of guake. The height
        is the window_height property, width is window_width and the
        horizontal alignment is given by window_alignment.
        """
        # fetch settings
        height_percents = settings.general.get_int("window-height")
        width_percents = settings.general.get_int("window-width")
        halignment = settings.general.get_int("window-halignment")
        valignment = settings.general.get_int("window-valignment")
        vdisplacement = settings.general.get_int("window-vertical-displacement")
        hdisplacement = settings.general.get_int("window-horizontal-displacement")

        log.debug("set_final_window_rect")
        log.debug("  height_percents = %s", height_percents)
        log.debug("  width_percents = %s", width_percents)
        log.debug("  halignment = %s", halignment)
        log.debug("  valignment = %s", valignment)
        log.debug("  hdisplacement = %s", hdisplacement)
        log.debug("  vdisplacement = %s", vdisplacement)

        # get the rectangle just from the destination monitor
        monitor = cls.get_final_window_monitor(settings, window)
        window_rect = monitor.get_workarea()
        log.debug("Current monitor geometry")
        log.debug("  window_rect.x: %s", window_rect.x)
        log.debug("  window_rect.y: %s", window_rect.y)
        log.debug("  window_rect.height: %s", window_rect.height)
        log.debug("  window_rect.width: %s", window_rect.width)

        total_height = window_rect.height
        total_width = window_rect.width

        if halignment == ALIGN_CENTER:
            log.debug("aligning to center!")
            window_rect.width = int(float(total_width) * float(width_percents) / 100.0)
            window_rect.x += (total_width - window_rect.width) / 2
        elif halignment == ALIGN_LEFT:
            log.debug("aligning to left!")
            window_rect.width = int(
                float(total_width - hdisplacement) * float(width_percents) / 100.0
            )
            window_rect.x += hdisplacement
        elif halignment == ALIGN_RIGHT:
            log.debug("aligning to right!")
            window_rect.width = int(
                float(total_width - hdisplacement) * float(width_percents) / 100.0
            )
            window_rect.x += total_width - window_rect.width - hdisplacement

        window_rect.height = int(float(total_height) * float(height_percents) / 100.0)
        if valignment == ALIGN_TOP:
            window_rect.y += vdisplacement
        elif valignment == ALIGN_BOTTOM:
            window_rect.y += total_height - window_rect.height - vdisplacement

        if width_percents == 100 and height_percents == 100:
            log.debug("MAXIMIZING MAIN WINDOW")
            window.move(window_rect.x, window_rect.y)
            window.maximize()
        elif not FullscreenManager(settings, window).is_fullscreen():
            log.debug("RESIZING MAIN WINDOW TO THE FOLLOWING VALUES:")
            window.unmaximize()
            log.debug("  window_rect.x: %s", window_rect.x)
            log.debug("  window_rect.y: %s", window_rect.y)
            log.debug("  window_rect.height: %s", window_rect.height)
            log.debug("  window_rect.width: %s", window_rect.width)
            # Note: move_resize is only on GTK3
            window.resize(window_rect.width, window_rect.height)
            window.move(window_rect.x, window_rect.y)
            log.debug("Updated window position: %r", window.get_position())

        return window_rect

    @classmethod
    def get_final_window_monitor(cls, settings, window):
        """Gets the final monitor for the main window of guake."""

        display = window.get_display()

        # fetch settings
        use_mouse = settings.general.get_boolean("mouse-display")
        num_monitor = settings.general.get_int("display-n")

        if use_mouse:
            pointer = display.get_default_seat().get_pointer()
            if pointer is None:
                monitor = display.get_primary_monitor()
            else:
                _, x, y = pointer.get_position()
                monitor = display.get_monitor_at_point(x, y)
        else:
            monitor = display.get_monitor(num_monitor)
            if monitor is None:
                # monitor not found or num_monitor is wrong
                # by default we use the primary monitor
                monitor = display.get_primary_monitor()

        return monitor


class ImageLayoutMode(enum.IntEnum):
    SCALE = 0
    TILE = 1
    CENTER = 2
    STRETCH = 3


class BackgroundImageManager:
    def __init__(self, window, filename=None, layout_mode=ImageLayoutMode.SCALE):
        self.window = window
        self.filename = ""
        self.bg_surface = self.load_from_file(filename) if filename else None
        self.target_surface = None
        self.target_info = (-1, -1, -1)  # (width, height, model)
        self._layout_mode = layout_mode

    @property
    def layout_mode(self):
        return self._layout_mode

    @layout_mode.setter
    def layout_mode(self, mode):
        mode = ImageLayoutMode(mode)
        if mode not in ImageLayoutMode:
            raise ValueError("Unknown layout mode")
        self._layout_mode = mode
        self.window.queue_draw()

    def load_from_file(self, filename):
        if not filename:
            # Clear the background image
            self.bg_surface = None
            self.window.queue_draw()
            return

        if not os.path.exists(filename):
            raise FileNotFoundError(f"Background file not found: {filename}")

        if self.filename:
            # Cached rendered surface
            if os.path.samefile(self.filename, filename):
                return self.bg_surface

        self.filename = filename
        img = Gtk.Image.new_from_file(filename)
        pixbuf = img.get_pixbuf()
        surface = cairo.ImageSurface(cairo.FORMAT_ARGB32, pixbuf.get_width(), pixbuf.get_height())
        cr = cairo.Context(surface)
        Gdk.cairo_set_source_pixbuf(cr, pixbuf, 0, 0)
        cr.set_operator(cairo.OPERATOR_SOURCE)
        cr.paint()

        self.bg_surface = surface
        self.target_info = (-1, -1, -1)
        self.window.queue_draw()
        return surface

    def render_target(self, width, height, mode, scale_mode=cairo.FILTER_BILINEAR):
        """Paint bacground image to the specific size target surface with different layout mode"""
        if not self.bg_surface:
            return None

        # Check if target surface has been rendered
        if self.target_info == (width, height, mode):
            return self.target_surface

        # Render new target
        self.target_info = (width, height, mode)
        surface = cairo.ImageSurface(cairo.FORMAT_ARGB32, width, height)
        cr = cairo.Context(surface)
        cr.set_operator(cairo.OPERATOR_SOURCE)

        xscale = width / self.bg_surface.get_width()
        yscale = height / self.bg_surface.get_height()
        if mode == ImageLayoutMode.SCALE:
            ratio = max(xscale, yscale)
            xoffset = (width - (self.bg_surface.get_width() * ratio)) / 2.0
            yoffset = (height - (self.bg_surface.get_height() * ratio)) / 2.0
            cr.translate(xoffset, yoffset)
            cr.scale(ratio, ratio)
            cr.set_source_surface(self.bg_surface, 0, 0)
            cr.get_source().set_filter(scale_mode)
        elif mode == ImageLayoutMode.TILE:
            cr.set_source_surface(self.bg_surface, 0, 0)
            cr.get_source().set_extend(cairo.EXTEND_REPEAT)
        elif mode == ImageLayoutMode.CENTER:
            x = (width - self.bg_surface.get_width()) / 2.0
            y = (height - self.bg_surface.get_height()) / 2.0
            cr.translate(x, y)
            cr.set_source_surface(self.bg_surface, 0, 0)
        elif mode == ImageLayoutMode.STRETCH:
            cr.scale(xscale, yscale)
            cr.set_source_surface(self.bg_surface, 0, 0)
            cr.get_source().set_filter(scale_mode)
        cr.paint()

        self.target_surface = surface
        return surface

    def draw(self, widget, cr):
        if not self.bg_surface:
            return

        # Step 1. Get target surface
        #         (paint background image into widget size surface by layout mode)
        surface = self.render_target(
            widget.get_allocated_width(),
            widget.get_allocated_height(),
            self.layout_mode,
        )

        cr.save()
        # Step 2. Paint target surface to context (in our case, the RootTerminalBox)
        #
        #         At this step, RootTerminalBox will be painted to our background image,
        #         and all the draw done by VTE has been overlapped.
        #
        cr.set_source_surface(surface, 0, 0)
        cr.paint()

        # Step 3. Re-paint child draw result to a new surface
        #
        #         We need to re-paint what we overlapped in previous step,
        #         so we paint our child (again, in lifecycle view) into a similar surface.
        #
        child = widget.get_child()
        child_surface = cr.get_target().create_similar(
            cairo.CONTENT_COLOR_ALPHA,
            child.get_allocated_width(),
            child.get_allocated_height(),
        )
        child_cr = cairo.Context(child_surface)

        # Re-paint child draw into child context (which using child_surface as target)
        widget.propagate_draw(child, child_cr)

        # Step 3.1 Re-paint search revealer
        child_sr_surface = None
        if getattr(widget, "search_revealer", None):
            child_sr = widget.search_revealer
            child_sr_surface = cr.get_target().create_similar(
                cairo.CONTENT_COLOR_ALPHA,
                child_sr.get_allocated_width(),
                child_sr.get_allocated_height(),
            )
            child_sr_cr = cairo.Context(child_surface)

            # Re-paint child draw into child context (which using child_surface as target)
            widget.propagate_draw(child_sr, child_sr_cr)

        # Step 4. Paint child surface into our context (RootTerminalBox)
        #
        #         Before this step, we have two important context/surface
        #             1. cr       / RootTerminalBox
        #             2. child_cr / child (RootTerminalBox's child)
        #         And current context have these draw:
        #             1. cr       - background image
        #             2. child_cr - child re-paint (split terminal, VTE draw ...etc)
        #         In this step, we are going to paint child_cr result back to cr,
        #         so in the end, we will get background image + other child stuff
        #
        # DEBUG: If you don't believe, use cairo.Surface.wrte_to_png(filename) to check what is
        #        inside the surface.
        #
        cr.set_source_surface(child_surface, 0, 0)
        cr.set_operator(cairo.OPERATOR_OVER)
        cr.paint()

        # Paint search revealer
        if child_sr_surface:
            cr.set_source_surface(child_sr_surface, 0, 0)
            cr.set_operator(cairo.OPERATOR_OVER)
            cr.paint()

        cr.restore()


def get_process_name(pid):
    stat_file = f"/proc/{pid}/stat"
    try:
        with open(stat_file, "r", encoding="utf-8") as fp:
            status = fp.read()
    except IOError as ex:
        log.debug("Unable to read %s: %s", stat_file, ex)
        status = ""

    match = re.match(r"\d+ \(([^)]+)\)", status)
    return match.group(1) if match else None



# -*- coding: utf-8; -*-
"""
Copyright (C) 2007-2018 Guake authors

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License as
published by the Free Software Foundation; either version 2 of the
License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
General Public License for more details.

You should have received a copy of the GNU General Public
License along with this program; if not, write to the
Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
Boston, MA 02110-1301 USA
"""

from guake.about import AboutDialog
from guake.boxes import RootTerminalBox
from guake.boxes import TabLabelEventBox
from guake.boxes import TerminalBox
from guake.callbacks import MenuHideCallback
from guake.callbacks import NotebookScrollCallback
from guake.dialogs import PromptQuitDialog
from guake.globals import PROMPT_ALWAYS
from guake.globals import PROMPT_PROCESSES
from guake.menus import mk_notebook_context_menu
from guake.prefs import PrefsDialog
from guake.utils import HidePrevention
from guake.utils import gdk_is_x11_display
from guake.utils import get_process_name
from guake.utils import save_tabs_when_changed

import gi
import os

gi.require_version("Gtk", "3.0")
gi.require_version("Wnck", "3.0")
from gi.repository import GObject
from gi.repository import Gdk
from gi.repository import Gtk
from gi.repository import Wnck
from guake.terminal import GuakeTerminal

import logging
import posix

log = logging.getLogger(__name__)


class TerminalNotebook(Gtk.Notebook):
    def __init__(self, *args, **kwargs):
        Gtk.Notebook.__init__(self, *args, **kwargs)
        self.last_terminal_focused = None

        self.set_name("notebook-teminals")
        self.set_tab_pos(Gtk.PositionType.BOTTOM)
        self.set_property("show-tabs", True)
        self.set_property("enable-popup", False)
        self.set_property("scrollable", True)
        self.set_property("show-border", False)
        self.set_property("visible", True)
        self.set_property("has-focus", True)
        self.set_property("can-focus", True)
        self.set_property("is-focus", True)
        self.set_property("expand", True)

        if GObject.signal_lookup("terminal-spawned", TerminalNotebook) == 0:
            GObject.signal_new(
                "terminal-spawned",
                TerminalNotebook,
                GObject.SignalFlags.RUN_LAST,
                GObject.TYPE_NONE,
                (GObject.TYPE_PYOBJECT, GObject.TYPE_INT),
            )
            GObject.signal_new(
                "page-deleted",
                TerminalNotebook,
                GObject.SignalFlags.RUN_LAST,
                GObject.TYPE_NONE,
                (),
            )

        self.scroll_callback = NotebookScrollCallback(self)
        self.add_events(Gdk.EventMask.SCROLL_MASK)
        self.connect("scroll-event", self.scroll_callback.on_scroll)
        self.notebook_on_button_press_id = self.connect(
            "button-press-event", self.on_button_press, None
        )

        # Action box
        self.pin_button = Gtk.ToggleButton(
            image=Gtk.Image.new_from_icon_name("view-pin-symbolic", Gtk.IconSize.MENU),
            visible=False,
        )
        self.pin_button.connect("clicked", self.on_pin_clicked)
        self.new_page_button = Gtk.Button(
            image=Gtk.Image.new_from_icon_name("tab-new-symbolic", Gtk.IconSize.MENU),
            visible=True,
        )
        self.new_page_button.connect("clicked", self.on_new_tab)

        self.tab_selection_button = Gtk.Button(
            image=Gtk.Image.new_from_icon_name("pan-down-symbolic", Gtk.IconSize.MENU),
            visible=True,
        )
        self.popover = Gtk.Popover()
        self.popover_window = None
        self.tab_selection_button.connect("clicked", self.on_tab_selection)

        self.action_box = Gtk.Box(visible=True)
        self.action_box.pack_start(self.pin_button, 0, 0, 0)
        self.action_box.pack_start(self.new_page_button, 0, 0, 0)
        self.action_box.pack_start(self.tab_selection_button, 0, 0, 0)
        self.set_action_widget(self.action_box, Gtk.PackType.END)

    def attach_guake(self, guake):
        self.guake = guake

        self.guake.settings.general.onChangedValue("window-losefocus", self.on_lose_focus_toggled)
        self.pin_button.set_visible(self.guake.settings.general.get_boolean("window-losefocus"))

    def on_button_press(self, target, event, user_data):
        if event.button == 3:
            menu = mk_notebook_context_menu(self)
            menu.connect("hide", MenuHideCallback(self.guake.window).on_hide)

            try:
                menu.popup_at_pointer(event)
            except AttributeError:
                # Gtk 3.18 fallback ("'Menu' object has no attribute 'popup_at_pointer'")
                menu.popup(None, None, None, None, event.button, event.time)
        elif (
            event.type == Gdk.EventType.DOUBLE_BUTTON_PRESS
            and event.button == 1
            and event.window.get_height() < 60
        ):
            # event.window.get_height() reports the height of the clicked frame
            self.new_page_with_focus()

        return False

    def on_pin_clicked(self, user_data=None):
        hide_prevention = HidePrevention(self.guake.window)
        if self.pin_button.get_active():
            hide_prevention.prevent()
        else:
            hide_prevention.allow()

    def on_lose_focus_toggled(self, settings, key, user_data=None):
        self.pin_button.set_visible(settings.get_boolean(key))

    @save_tabs_when_changed
    def on_new_tab(self, user_data):
        self.new_page_with_focus()

    def on_tab_selection(self, user_data):
        """Construct the tab selection popover

        Since we did not use Gtk.ListStore to store tab information, we will construct the
        tab selection popover content each time when user click them.
        """

        # Remove previous window
        if self.popover_window:
            self.popover.remove(self.popover_window)

        # This makes the list's background transparent
        # ref: epiphany
        css_provider = Gtk.CssProvider()
        css_provider.load_from_data(
            b"#popover-window list { border-style: none; background-color: transparent; }"
        )
        Gtk.StyleContext.add_provider_for_screen(
            Gdk.Screen.get_default(),
            css_provider,
            Gtk.STYLE_PROVIDER_PRIORITY_APPLICATION,
        )

        # Construct popover properties
        BOX_HEIGHT = 30
        LISTBOX_MARGIN = 12
        self.popover_window = Gtk.ScrolledWindow(name="popover-window")
        self.popover_listbox = Gtk.ListBox()
        self.popover_listbox.set_property("margin", LISTBOX_MARGIN)
        self.popover_window.add_with_viewport(self.popover_listbox)

        max_height = (
            self.guake.window.get_allocation().height - BOX_HEIGHT
            if self.guake
            else BOX_HEIGHT * 10
        )
        height = BOX_HEIGHT * self.get_n_pages() + LISTBOX_MARGIN * 4
        self.popover_window.set_min_content_height(min(max_height, height))
        self.popover_window.set_min_content_width(325)
        self.popover.add(self.popover_window)

        # Construct content
        current_term = self.get_current_terminal()
        selected_row = 0
        for i in range(self.get_n_pages()):
            row = Gtk.ListBoxRow()
            box = Gtk.Box(orientation=Gtk.Orientation.HORIZONTAL)
            box.set_size_request(200, BOX_HEIGHT)
            label = Gtk.Label(self.get_tab_text_index(i))
            label.set_xalign(0.0)
            box.pack_start(label, 0, 0, 5)
            row.add(box)
            setattr(row, "page_index", i)
            self.popover_listbox.add(row)
            if current_term in self.get_terminals_for_page(i):
                self.popover_listbox.select_row(row)
                selected_row = i

        # Signal
        self.popover_listbox.connect("row-activated", self.on_popover_tab_select)

        # Show popup
        self.popover.set_position(Gtk.PositionType.TOP)
        self.popover.set_relative_to(user_data)
        self.popover.show_all()
        try:
            # For GTK >= 3.22
            self.popover.popup()
        except AttributeError:
            pass

        # Adjust scrollor
        while Gtk.events_pending():
            Gtk.main_iteration()

        if selected_row:
            adj = self.popover_window.get_vadjustment()
            v = adj.get_upper() - adj.get_page_size()
            part = v / self.get_n_pages()
            adj.set_value(part * (selected_row + 1))

    def on_popover_tab_select(self, list_box, row):
        page_index = getattr(row, "page_index", -1)
        if page_index != -1:
            self.set_current_page(page_index)
            self.get_terminals_for_page(page_index)[0].grab_focus()

    def set_tabbar_visible(self, v):
        self.set_property("show-tabs", v)

    def set_last_terminal_focused(self, terminal):
        self.last_terminal_focused = terminal

    def get_focused_terminal(self):
        for terminal in self.iter_terminals():
            if terminal.has_focus():
                return terminal

    def get_current_terminal(self):
        # TODO NOTEBOOK the name of this method should
        # be changed, for now it returns the last focused terminal
        return self.last_terminal_focused

    def get_terminals_for_page(self, index):
        page = self.get_nth_page(index)
        return page.get_terminals()

    def get_terminals(self):
        terminals = []
        for page in self.iter_pages():
            terminals += page.get_terminals()
        return terminals

    def get_running_fg_processes(self):
        processes = []
        for page in self.iter_pages():
            processes += self.get_running_fg_processes_page(page)
        return processes

    def get_running_fg_processes_page(self, page):
        processes = []
        for terminal in page.get_terminals():
            pty = terminal.get_pty()
            if not pty:
                continue
            fdpty = pty.get_fd()
            term_pid = terminal.pid
            try:
                fgpid = posix.tcgetpgrp(fdpty)
                log.debug("found running pid: %s", fgpid)
                if fgpid not in (-1, term_pid):
                    processes.append((fgpid, get_process_name(fgpid)))
            except OSError:
                log.debug(
                    "Cannot retrieve any pid from terminal %s, looks like it is already dead",
                    terminal,
                )
        return processes

    def has_page(self):
        return self.get_n_pages() > 0

    def iter_terminals(self):
        for page in self.iter_pages():
            if page is not None:
                for t in page.iter_terminals():
                    yield t

    def iter_tabs(self):
        for page_num in range(self.get_n_pages()):
            yield self.get_tab_label(self.get_nth_page(page_num))

    def iter_pages(self):
        for page_num in range(self.get_n_pages()):
            yield self.get_nth_page(page_num)

    def delete_page(self, page_num, kill=True, prompt=0):
        log.debug("Deleting page index %s", page_num)
        if page_num >= self.get_n_pages() or page_num < 0:
            log.error("Can not delete page %s no such index", page_num)
            return

        page = self.get_nth_page(page_num)
        # TODO NOTEBOOK it would be nice if none of the "ui" stuff
        # (PromptQuitDialog) would be in here
        procs = self.get_running_fg_processes_page(page)
        if prompt == PROMPT_ALWAYS or (prompt == PROMPT_PROCESSES and procs):
            # TODO NOTEBOOK remove call to guake
            if not PromptQuitDialog(self.guake.window, procs, -1, None).close_tab():
                return

        for terminal in page.get_terminals():
            if kill:
                terminal.kill()
            terminal.destroy()

        if self.get_nth_page(page_num) is page:
            # NOTE: GitHub issue #1438
            # Previous line `terminal.destroy()` will finally called `on_terminal_exited`,
            # and called `RootTerminalBox.remove_dead_child`, then called `remove_page`.
            #
            # But in some cases (e.g. #1438), it will not remove the page by
            # `terminal.destory() chain`.
            #
            # Check this by compare same page_num page with previous saved page instance,
            # and remove the page if it really didn't remove it.
            self.remove_page(page_num)

    @save_tabs_when_changed
    def remove_page(self, page_num):
        super().remove_page(page_num)
        # focusing the first terminal on the previous page
        if self.get_current_page() > -1:
            page = self.get_nth_page(self.get_current_page())
            if page.get_terminals():
                page.get_terminals()[0].grab_focus()

        self.hide_tabbar_if_one_tab()
        self.emit("page-deleted")

    def delete_page_by_label(self, label, kill=True, prompt=0):
        self.delete_page(self.find_tab_index_by_label(label), kill, prompt)

    def delete_page_current(self, kill=True, prompt=0):
        self.delete_page(self.get_current_page(), kill, prompt)

    def new_page(self, directory=None, position=None, empty=False, open_tab_cwd=False):
        terminal_box = TerminalBox()
        if empty:
            terminal = None
        else:
            terminal = self.terminal_spawn(directory, open_tab_cwd)
            terminal_box.set_terminal(terminal)
        root_terminal_box = RootTerminalBox(self.guake, self)
        root_terminal_box.set_child(terminal_box)
        page_num = self.insert_page(
            root_terminal_box, None, position if position is not None else -1
        )
        self.set_tab_reorderable(root_terminal_box, True)
        self.show_all()  # needed to show newly added tabs and pages
        # this is needed because self.window.show_all() results in showing every
        # thing which includes the scrollbar too
        self.guake.settings.general.triggerOnChangedValue(
            self.guake.settings.general, "use-scrollbar"
        )
        # this is needed to initially set the last_terminal_focused,
        # one could also call terminal.get_parent().on_terminal_focus()
        if not empty:
            self.terminal_attached(terminal)
        self.hide_tabbar_if_one_tab()

        if self.guake:
            # Attack background image draw callback to root terminal box
            root_terminal_box.connect_after("draw", self.guake.background_image_manager.draw)
        return root_terminal_box, page_num, terminal

    def hide_tabbar_if_one_tab(self):
        """Hide the tab bar if hide-tabs-if-one-tab is true and there is only one
        notebook page"""
        if self.guake.settings.general.get_boolean("window-tabbar"):
            if self.guake.settings.general.get_boolean("hide-tabs-if-one-tab"):
                self.set_property("show-tabs", self.get_n_pages() != 1)
            else:
                self.set_property("show-tabs", True)

    def terminal_spawn(self, directory=None, open_tab_cwd=False):
        terminal = GuakeTerminal(self.guake)
        terminal.grab_focus()
        terminal.connect(
            "key-press-event",
            lambda x, y: self.guake.accel_group.activate(x, y) if self.guake.accel_group else False,
        )
        if not isinstance(directory, str):
            directory = os.environ["HOME"]
            try:
                if self.guake.settings.general.get_boolean("open-tab-cwd") or open_tab_cwd:
                    # Do last focused terminal still alive?
                    active_terminal = self.get_current_terminal()
                    if not active_terminal:
                        # If not alive, can we get any focused terminal?
                        active_terminal = self.get_focused_terminal()
                    directory = os.path.expanduser("~")
                    if active_terminal:
                        # If found, we will use its directory as new terminal's directory
                        directory = active_terminal.get_current_directory()
            except BaseException:
                pass
        log.info("Spawning new terminal at %s", directory)
        terminal.spawn_sync_pid(directory)
        return terminal

    def terminal_attached(self, terminal):
        terminal.emit("focus", Gtk.DirectionType.TAB_FORWARD)
        self.emit("terminal-spawned", terminal, terminal.pid)

    def new_page_with_focus(
        self,
        directory=None,
        label=None,
        user_set=False,
        position=None,
        empty=False,
        open_tab_cwd=False,
    ):
        box, page_num, terminal = self.new_page(
            directory, position=position, empty=empty, open_tab_cwd=open_tab_cwd
        )
        self.set_current_page(page_num)
        if not label:
            self.rename_page(page_num, self.guake.compute_tab_title(terminal), False)
        else:
            self.rename_page(page_num, label, user_set)
        if terminal is not None:
            terminal.grab_focus()
        return box, page_num, terminal

    def rename_page(self, page_index, new_text, user_set=False):
        """Rename an already added page by its index. Use user_set to define
        if the rename was triggered by the user (eg. rename dialog) or by
        an update from the vte (eg. vte:window-title-changed)
        """
        page = self.get_nth_page(page_index)
        if not getattr(page, "custom_label_set", False) or user_set:
            old_label = self.get_tab_label(page)
            if isinstance(old_label, TabLabelEventBox):
                old_label.set_text(new_text)
            else:
                label = TabLabelEventBox(self, new_text, self.guake.settings)
                label.add_events(Gdk.EventMask.SCROLL_MASK)
                label.connect("scroll-event", self.scroll_callback.on_scroll)

                self.set_tab_label(page, label)
            if user_set:
                setattr(page, "custom_label_set", new_text != "-")

    def find_tab_index_by_label(self, eventbox):
        for index, tab_eventbox in enumerate(self.iter_tabs()):
            if eventbox is tab_eventbox:
                return index
        return -1

    def find_page_index_by_terminal(self, terminal):
        for index, page in enumerate(self.iter_pages()):
            for t in page.iter_terminals():
                if t is terminal:
                    return index
        return -1

    def get_tab_text_index(self, index):
        return self.get_tab_label(self.get_nth_page(index)).get_text()

    def get_tab_text_page(self, page):
        return self.get_tab_label(page).get_text()

    def on_show_preferences(self, user_data):
        self.guake.hide()
        PrefsDialog(self.guake.settings).show()

    def on_show_about(self, user_data):
        self.guake.hide()
        AboutDialog()

    def on_quit(self, user_data):
        self.guake.accel_quit()

    def on_save_tabs(self, user_data):
        self.guake.save_tabs()

    def on_restore_tabs(self, user_data):
        self.guake.restore_tabs()

    def on_restore_tabs_with_dialog(self, user_data):
        dialog = Gtk.MessageDialog(
            parent=self.guake.window,
            flags=Gtk.DialogFlags.MODAL,
            buttons=Gtk.ButtonsType.OK_CANCEL,
            message_format=_(
                "You are going to restore *all* the tabs!\n"
                "which means all your terminals & pages "
                "will be replaced.\n\nDo you want to continue?"
            ),
        )
        dialog.connect("response", self.restore_tabs_dialog_response)
        dialog.show()

    def restore_tabs_dialog_response(self, widget, response_id):
        widget.destroy()
        if response_id == Gtk.ResponseType.OK:
            self.guake.restore_tabs()


class NotebookManager(GObject.Object):
    def __init__(
        self,
        window,
        notebook_parent,
        workspaces_enabled,
        terminal_spawned_cb,
        page_deleted_cb,
    ):
        GObject.Object.__init__(self)
        if not GObject.signal_lookup("notebook-created", self):
            GObject.signal_new(
                "notebook-created",
                self,
                GObject.SignalFlags.RUN_LAST,
                GObject.TYPE_NONE,
                (GObject.TYPE_PYOBJECT, GObject.TYPE_INT),
            )
        self.current_notebook = 0
        self.notebooks = {}
        self.window = window
        self.notebook_parent = notebook_parent
        self.terminal_spawned_cb = terminal_spawned_cb
        self.page_deleted_cb = page_deleted_cb
        if workspaces_enabled and gdk_is_x11_display(Gdk.Display.get_default()):
            # NOTE: Wnck didn't support non-X11 display backend, so we need to check if the display
            #       is X11 or not, if not, it will not able to enable workspace-specific-tab-sets
            #
            # TODO: Is there anyway to support this in non-X11 display backend?
            self.screen = Wnck.Screen.get_default()
            self.screen.connect("active-workspace-changed", self.__workspace_changed_cb)

    def __workspace_changed_cb(self, screen, previous_workspace):
        self.set_workspace(self.screen.get_active_workspace().get_number())

    def get_notebook(self, workspace_index: int):
        if not self.has_notebook_for_workspace(workspace_index):
            self.notebooks[workspace_index] = TerminalNotebook()
            self.emit("notebook-created", self.notebooks[workspace_index], workspace_index)
            self.notebooks[workspace_index].connect("terminal-spawned", self.terminal_spawned_cb)
            self.notebooks[workspace_index].connect("page-deleted", self.page_deleted_cb)
            log.info("created fresh notebook for workspace %d", self.current_notebook)

            # add a tab if there is none
            if not self.notebooks[workspace_index].has_page():
                self.notebooks[workspace_index].new_page_with_focus(None)

        return self.notebooks[workspace_index]

    def get_current_notebook(self):
        return self.get_notebook(self.current_notebook)

    def has_notebook_for_workspace(self, workspace_index):
        return workspace_index in self.notebooks

    def set_workspace(self, index: int):
        self.notebook_parent.remove(self.get_current_notebook())
        self.current_notebook = index
        log.info("current workspace is %d", self.current_notebook)
        notebook = self.get_current_notebook()
        self.notebook_parent.add(notebook)
        if self.window.get_property("visible") and notebook.last_terminal_focused is not None:
            notebook.last_terminal_focused.grab_focus()

        # Restore pending page terminal split
        notebook.guake.restore_pending_terminal_split()

        # Restore config to workspace
        notebook.guake.load_config()

    def set_notebooks_tabbar_visible(self, v):
        for nb in self.iter_notebooks():
            nb.set_tabbar_visible(v)

    def get_notebooks(self):
        return self.notebooks

    def get_terminals(self):
        terminals = []
        for k in self.notebooks:
            terminals += self.notebooks[k].get_terminals()
        return terminals

    def iter_terminals(self):
        for k in self.notebooks:
            for t in self.notebooks[k].iter_terminals():
                yield t

    def get_terminal_by_uuid(self, terminal_uuid):
        for t in self.iter_terminals():
            if t.uuid == terminal_uuid:
                return t
        return None

    def iter_pages(self):
        for k in self.notebooks:
            for t in self.notebooks[k].iter_pages():
                yield t

    def iter_notebooks(self):
        for k in self.notebooks:
            yield self.notebooks[k]

    def get_n_pages(self):
        n = 0
        for k in self.notebooks:
            n += self.notebooks[k].get_n_pages()
        return n

    def get_n_notebooks(self):
        return len(self.notebooks.keys())

    def get_running_fg_processes(self):
        processes = []
        for k in self.notebooks:
            processes += self.notebooks[k].get_running_fg_processes()
        return processes



# -*- coding: utf-8; -*-
"""
Copyright (C) 2007-2013 Guake authors

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License as
published by the Free Software Foundation; either version 2 of the
License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
General Public License for more details.

You should have received a copy of the GNU General Public
License along with this program; if not, write to the
Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
Boston, MA 02110-1301 USA
"""

import inspect
import logging
import os

__all__ = [
    "ALIGN_BOTTOM",
    "ALIGN_CENTER",
    "ALIGN_LEFT",
    "ALIGN_RIGHT",
    "ALIGN_TOP",
    "ALWAYS_ON_PRIMARY",
    "NAME",
]

log = logging.getLogger(__name__)


def bindtextdomain(app_name, locale_dir=None):
    """
    Bind the domain represented by app_name to the locale directory locale_dir.
    It has the effect of loading translations, enabling applications for different
    languages.

    app_name:
        a domain to look for translations, typically the name of an application.

    locale_dir:
        a directory with locales like locale_dir/lang_isocode/LC_MESSAGES/app_name.mo
        If omitted or None, then the current binding for app_name is used.
    """

    # pylint: disable=import-outside-toplevel
    import locale

    # pylint: enable=import-outside-toplevel

    log.info("Local binding for app '%s', local dir: %s", app_name, locale_dir)

    locale.bindtextdomain(app_name, locale_dir)
    locale.textdomain(app_name)


def is_run_from_git_workdir():
    self_path = os.path.abspath(inspect.getfile(inspect.currentframe()))
    return os.path.exists(f"{self_path}.in")


NAME = "guake"

ALIGN_CENTER, ALIGN_LEFT, ALIGN_RIGHT = range(3)
ALIGN_TOP, ALIGN_BOTTOM = range(2)
ALWAYS_ON_PRIMARY = -1
PROMPT_NEVER, PROMPT_PROCESSES, PROMPT_ALWAYS = range(3)

# TODO this is not as fancy as as it could be
# pylint: disable=anomalous-backslash-in-string
TERMINAL_MATCH_TAGS = ("schema", "http", "https", "email", "ftp")
# Beware this is a PRCE (Perl) regular expression, not a Python one!
# Edit: use regex101.com with PCRE syntax
TERMINAL_MATCH_EXPRS = [
    r"(news:|telnet:|nntp:|file:\/|https?:|ftps?:|webcal:)\/\/([-[:alnum:]]+"
    r"(:[-[:alnum:],?;.:\/!%$^\*&~\"#']+)?\@)?[-[:alnum:]]+(\.[-[:alnum:]]+)*"
    r"(:[0-9]{1,5})?(\/[-[:alnum:]_$.+!*(),;:@&=?\/~#%]*[^]'.>) \t\r\n,\\\"])?",
    r"(www|ftp)[-[:alnum:]]*\.[-[:alnum:]]+(\.[-[:alnum:]]+)*(:[0-9]{1,5})?"
    r"(\/[-[:alnum:]_$.+!*(),;:@&=?\/~#%]*[^]'.>) \t\r\n,\\\"])?",
    r"(mailto:)?[-[:alnum:]][-[:alnum:].]*@[-[:alnum:]]+\.[-[:alnum:]]+(\\.[-[:alnum:]]+)*",
]
# tuple (title/quick matcher/filename and line number extractor)
QUICK_OPEN_MATCHERS = [
    (
        "Python traceback",
        r"^\s*File\s\".*\",\sline\s[0-9]+",
        r"^\s*File\s\"(.*)\",\sline\s([0-9]+)",
    ),
    (
        "Python pytest report",
        r"^\s.*\:\:[a-zA-Z0-9\_]+\s",
        r"^\s*(.*\:\:[a-zA-Z0-9\_]+)\s",
    ),
    (
        "line starts by 'ERROR in Filename:line' pattern (GCC/make). File path should exists.",
        r"[a-zA-Z0-9\/\_\-\.\]+\.?[a-zA-Z0-9]+\:[0-9]+",
        r"\s.\S[^\s\s].(.*)\:([0-9]+)",
    ),
    (
        "line starts by 'Filename:line' pattern (GCC/make). File path should exists.",
        r"^\s*[a-zA-Z0-9\/\_\-\.\ ]+\.?[a-zA-Z0-9]+\:[0-9]+",
        r"^\s*(.*)\:([0-9]+)",
    ),
]

# Transparency max level (should be always 100)
MAX_TRANSPARENCY = 100

# Tabs session schema version
TABS_SESSION_SCHEMA_VERSION = 2

# Constants for vte regex matching are documented in the pcre2 api:
#   https://www.pcre.org/current/doc/html/pcre2api.html
PCRE2_MULTILINE = 0x00000400

# the urls of the search engine options for the search on web feature.
# Additional engines should be added
ENGINES = {
    0: "www.google.com/search?safe=off&q=",
    1: "www.duckduckgo.com/",
    2: "www.bing.com/search?q=",
    3: "www.yandex.com/search?text=",
    4: "neeva.com/search?q=",
}



import gi

gi.require_version("Gtk", "3.0")
from gi.repository import Gtk

from guake.customcommands import CustomCommands

import logging

log = logging.getLogger(__name__)


def mk_tab_context_menu(callback_object):
    """Create the context menu for a notebook tab"""
    # Store the menu in a temp variable in terminal so that popup() is happy. See:
    #   https://stackoverflow.com/questions/28465956/
    callback_object.context_menu = Gtk.Menu()
    menu = callback_object.context_menu
    mi_new_tab = Gtk.MenuItem(_("New Tab"))
    mi_new_tab.connect("activate", callback_object.on_new_tab)
    menu.add(mi_new_tab)
    mi_rename = Gtk.MenuItem(_("Rename"))
    mi_rename.connect("activate", callback_object.on_rename)
    menu.add(mi_rename)
    mi_reset_custom_colors = Gtk.MenuItem(_("Reset custom colors"))
    mi_reset_custom_colors.connect("activate", callback_object.on_reset_custom_colors)
    menu.add(mi_reset_custom_colors)
    mi_close = Gtk.MenuItem(_("Close"))
    mi_close.connect("activate", callback_object.on_close)
    menu.add(mi_close)
    menu.show_all()
    return menu


def mk_notebook_context_menu(callback_object):
    """Create the context menu for the notebook"""
    callback_object.context_menu = Gtk.Menu()
    menu = callback_object.context_menu
    mi = Gtk.MenuItem(_("New Tab"))
    mi.connect("activate", callback_object.on_new_tab)
    menu.add(mi)
    menu.add(Gtk.SeparatorMenuItem())
    mi = Gtk.MenuItem(_("Save Tabs"))
    mi.connect("activate", callback_object.on_save_tabs)
    menu.add(mi)
    mi = Gtk.MenuItem(_("Restore Tabs"))
    mi.connect("activate", callback_object.on_restore_tabs_with_dialog)
    menu.add(mi)
    menu.add(Gtk.SeparatorMenuItem())
    mi = Gtk.ImageMenuItem("gtk-preferences")
    mi.set_use_stock(True)
    mi.connect("activate", callback_object.on_show_preferences)
    menu.add(mi)
    mi = Gtk.ImageMenuItem("gtk-about")
    mi.set_use_stock(True)
    mi.connect("activate", callback_object.on_show_about)
    menu.add(mi)
    menu.add(Gtk.SeparatorMenuItem())
    mi = Gtk.MenuItem(_("Quit"))
    mi.connect("activate", callback_object.on_quit)
    menu.add(mi)
    menu.show_all()
    return menu


SEARCH_SELECTION_LENGTH = 20
FILE_SELECTION_LENGTH = 30


def mk_terminal_context_menu(terminal, window, settings, callback_object):
    """Create the context menu for a terminal."""
    # Store the menu in a temp variable in terminal so that popup() is happy. See:
    #   https://stackoverflow.com/questions/28465956/
    terminal.context_menu = Gtk.Menu()
    menu = terminal.context_menu
    mi = Gtk.MenuItem(_("Copy"))
    mi.connect("activate", callback_object.on_copy_clipboard)
    menu.add(mi)
    if get_link_under_cursor(terminal) is not None:
        mi = Gtk.MenuItem(_("Copy URL"))
        mi.connect("activate", callback_object.on_copy_url_clipboard)
        menu.add(mi)
    mi = Gtk.MenuItem(_("Paste"))
    mi.connect("activate", callback_object.on_paste_clipboard)
    # check if clipboard has text, if not disable the paste menuitem
    clipboard = Gtk.Clipboard.get_default(window.get_display())
    mi.set_sensitive(clipboard.wait_is_text_available())
    menu.add(mi)
    menu.add(Gtk.SeparatorMenuItem())
    mi = Gtk.MenuItem(_("Toggle Fullscreen"))
    mi.connect("activate", callback_object.on_toggle_fullscreen)
    menu.add(mi)
    menu.add(Gtk.SeparatorMenuItem())
    mi = Gtk.MenuItem(_("Split ―"))
    mi.connect("activate", callback_object.on_split_horizontal)
    menu.add(mi)
    mi = Gtk.MenuItem(_("Split |"))
    mi.connect("activate", callback_object.on_split_vertical)
    menu.add(mi)
    mi = Gtk.MenuItem(_("Close terminal"))
    mi.connect("activate", callback_object.on_close_terminal)
    menu.add(mi)
    menu.add(Gtk.SeparatorMenuItem())
    mi = Gtk.MenuItem(_("Save content..."))
    mi.connect("activate", callback_object.on_save_to_file)
    menu.add(mi)
    mi = Gtk.MenuItem(_("Reset terminal"))
    mi.connect("activate", callback_object.on_reset_terminal)
    menu.add(mi)
    # TODO SEARCH uncomment menu.add()
    mi = Gtk.MenuItem(_("Find..."))
    mi.connect("activate", callback_object.on_find)
    # menu.add(mi)
    menu.add(Gtk.SeparatorMenuItem())
    mi = Gtk.MenuItem(_("Open link..."))
    mi.connect("activate", callback_object.on_open_link)
    link = get_link_under_cursor(terminal)
    # TODO CONTEXTMENU this is a mess Quick open should also be sensible
    # if the text in the selection is a url the current terminal
    # implementation does not support this at the moment
    if link:
        if len(link) >= FILE_SELECTION_LENGTH:
            mi.set_label(_("Open Link: {!s}...").format(link[: FILE_SELECTION_LENGTH - 3]))
        else:
            mi.set_label(_("Open Link: {!s}").format(link))
        mi.set_sensitive(True)
    else:
        mi.set_sensitive(False)
    menu.add(mi)
    mi = Gtk.MenuItem(_("Search on Web"))
    mi.connect("activate", callback_object.on_search_on_web)
    selection = get_current_selection(terminal, window)
    if selection:
        search_text = selection.rstrip()
        if len(search_text) > SEARCH_SELECTION_LENGTH:
            search_text = search_text[: SEARCH_SELECTION_LENGTH - 3] + "..."
        mi.set_label(_("Search on Web: '%s'") % search_text)
        mi.set_sensitive(True)
    else:
        mi.set_sensitive(False)
    menu.add(mi)
    mi = Gtk.MenuItem(_("Quick Open..."))
    mi.connect("activate", callback_object.on_quick_open)
    if selection:
        filename = get_filename_under_cursor(terminal, selection)
        if filename:
            filename_str = str(filename)
            if len(filename_str) > FILE_SELECTION_LENGTH:
                mi.set_label(
                    _("Quick Open: {!s}...").format(filename_str[: FILE_SELECTION_LENGTH - 3])
                )
            else:
                mi.set_label(_("Quick Open: {!s}").format(filename_str))
            mi.set_sensitive(True)
        else:
            mi.set_sensitive(False)
    else:
        mi.set_sensitive(False)
    menu.add(mi)
    customcommands = CustomCommands(settings, callback_object)
    if customcommands.should_load():
        submen = customcommands.build_menu()
        if submen:
            menu.add(Gtk.SeparatorMenuItem())
            mi = Gtk.MenuItem(_("Custom Commands"))
            mi.set_submenu(submen)
            menu.add(mi)
    menu.add(Gtk.SeparatorMenuItem())
    mi = Gtk.ImageMenuItem("gtk-preferences")
    mi.set_use_stock(True)
    mi.connect("activate", callback_object.on_show_preferences)
    menu.add(mi)
    mi = Gtk.ImageMenuItem("gtk-about")
    mi.set_use_stock(True)
    mi.connect("activate", callback_object.on_show_about)
    menu.add(mi)
    menu.add(Gtk.SeparatorMenuItem())
    mi = Gtk.ImageMenuItem(_("Quit"))
    mi.connect("activate", callback_object.on_quit)
    menu.add(mi)
    menu.show_all()
    return menu


def get_current_selection(terminal, window):
    if terminal.get_has_selection():
        terminal.copy_clipboard()
        clipboard = Gtk.Clipboard.get_default(window.get_display())
        return clipboard.wait_for_text()
    return None


def get_filename_under_cursor(terminal, selection):
    filename, _1, _2 = terminal.is_file_on_local_server(selection)
    log.info("Current filename under cursor: %s", filename)
    if filename:
        return filename
    return None


def get_link_under_cursor(terminal):
    link = terminal.found_link
    log.info("Current link under cursor: %s", link)
    if link:
        return link
    return None



# -*- coding: utf-8; -*-
"""
Copyright (C) 2007-2013 Guake authors

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License as
published by the Free Software Foundation; either version 2 of the
License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
General Public License for more details.

You should have received a copy of the GNU General Public
License along with this program; if not, write to the
Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
Boston, MA 02110-1301 USA
"""

import logging
import logging.config

try:
    from colorlog import ColoredFormatter
except ImportError:
    ColoredFormatter = None

log = logging.getLogger(__name__)


def setupLogging(debug_mode):
    if debug_mode:
        base_logging_level = logging.DEBUG
    else:
        base_logging_level = logging.INFO

    if ColoredFormatter:
        level_str = logging.getLevelName(base_logging_level)
        logging.config.dictConfig(
            {
                "version": 1,
                "disable_existing_loggers": False,
                "loggers": {
                    "": {
                        "handlers": ["default"],
                        "level": level_str,
                        "propagate": True,
                    },
                },
                "handlers": {
                    "default": {
                        "level": level_str,
                        "class": "logging.StreamHandler",
                        "formatter": "default",
                    },
                },
                "formatters": {
                    "default": {
                        "()": "colorlog.ColoredFormatter",
                        "format": "%(log_color)s%(levelname)-8s%(reset)s %(message)s",
                        "log_colors": {
                            "DEBUG": "cyan",
                            "INFO": "green",
                            "WARNING": "yellow",
                            "ERROR": "red",
                            "CRITICAL": "red,bg_white",
                        },
                    }
                },
            }
        )
    else:
        logging.basicConfig(level=base_logging_level, format="%(message)s")
    log.setLevel(base_logging_level)
    log.debug("Logging configuration complete")



# -*- coding: utf-8; -*-
"""
Copyright (C) 2007-2012 Lincoln de Sousa <lincoln@minaslivre.org>
Copyright (C) 2007 Gabriel Falcão <gabrielteratos@gmail.com>

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License as
published by the Free Software Foundation; either version 2 of the
License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
General Public License for more details.

You should have received a copy of the GNU General Public
License along with this program; if not, write to the
Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
Boston, MA 02110-1301 USA
"""

import gi

gi.require_version("Gtk", "3.0")
from gi.repository import Gtk

from guake import guake_version
from guake.common import gladefile
from guake.common import pixmapfile
from guake.simplegladeapp import SimpleGladeApp


class AboutDialog(SimpleGladeApp):

    """The About Guake dialog class"""

    def __init__(self):
        super().__init__(gladefile("about.glade"), root="aboutdialog")
        dialog = self.get_widget("aboutdialog")

        # images
        image = Gtk.Image()
        image.set_from_file(pixmapfile("guake-notification.png"))
        pixbuf = image.get_pixbuf()

        dialog.set_property("logo", pixbuf)

        dialog.set_name(_("Guake Terminal"))
        dialog.set_version(guake_version())
        dialog.connect("response", lambda x, y: dialog.destroy())
